import os
import time
import unittest
import uuid

import pandas as pd
from kafka import KafkaProducer, KafkaConsumer
from kafka.admin import KafkaAdminClient, NewTopic

from feldera import PipelineBuilder, Pipeline
from feldera.enums import PipelineStatus
from tests import TEST_CLIENT, KAFKA_SERVER, PIPELINE_TO_KAFKA_SERVER


class TestPipelineBuilder(unittest.TestCase):
    def test_local(self):
        TBL_NAMES = ["students", "grades"]
        view_name = "average_scores"

        sql = f"""
        CREATE TABLE {TBL_NAMES[0]} (
            name STRING,
            id INT
        );

        CREATE TABLE {TBL_NAMES[1]} (
            student_id INT,
            science INT,
            maths INT,
            art INT
        );

        CREATE VIEW {view_name} AS SELECT name, ((science + maths + art) / 3) as average FROM {TBL_NAMES[0]} JOIN {TBL_NAMES[1]} on id = student_id ORDER BY average DESC;
        """

        pipeline = PipelineBuilder(
            TEST_CLIENT, name="notebook", sql=sql
        ).create_or_replace()

        df_students = pd.read_csv("tests/students.csv")
        df_grades = pd.read_csv("tests/grades.csv")

        out = pipeline.listen("average_scores")

        pipeline.start()
        pipeline.input_pandas(TBL_NAMES[0], df_students)
        pipeline.input_pandas(TBL_NAMES[1], df_grades)
        pipeline.wait_for_completion(True)
        df = out.to_pandas()

        assert df.shape[0] == 100

        pipeline.delete(True)

    def test_pipeline_get(self):
        TBL_NAMES = ["students", "grades"]
        view_name = "average_scores"

        sql = f"""
        CREATE TABLE {TBL_NAMES[0]} (
            name STRING,
            id INT
        );

        CREATE TABLE {TBL_NAMES[1]} (
            student_id INT,
            science INT,
            maths INT,
            art INT
        );

        CREATE VIEW {view_name} AS SELECT name, ((science + maths + art) / 3) as average FROM {TBL_NAMES[0]} JOIN {TBL_NAMES[1]} on id = student_id ORDER BY average DESC;
        """

        pipeline = PipelineBuilder(
            TEST_CLIENT, name="notebook", sql=sql
        ).create_or_replace()

        df_students = pd.read_csv("tests/students.csv")
        df_grades = pd.read_csv("tests/grades.csv")

        out = pipeline.listen("average_scores")

        pipeline.start()
        pipeline.input_pandas(TBL_NAMES[0], df_students)
        pipeline.input_pandas(TBL_NAMES[1], df_grades)
        pipeline.wait_for_completion(True)
        df = out.to_pandas()

        assert df.shape[0] == 100

        del pipeline
        del out

        pipeline = Pipeline.get("notebook", TEST_CLIENT)
        assert pipeline is not None
        pipeline.start()

        out = pipeline.listen("average_scores")
        pipeline.input_pandas(TBL_NAMES[0], df_students)
        pipeline.input_pandas(TBL_NAMES[1], df_grades)
        pipeline.wait_for_completion(True)
        df = out.to_pandas()

        assert df.shape[0] == 100
        pipeline.pause()
        pipeline.stop(force=True)
        pipeline.delete(True)

    def test_local_listen_after_start(self):
        TBL_NAMES = ["students", "grades"]
        view_name = "average_scores"

        sql = f"""
        CREATE TABLE {TBL_NAMES[0]} (
            name STRING,
            id INT
        );

        CREATE TABLE {TBL_NAMES[1]} (
            student_id INT,
            science INT,
            maths INT,
            art INT
        );

        CREATE VIEW {view_name} AS SELECT name, ((science + maths + art) / 3) as average FROM {TBL_NAMES[0]} JOIN {TBL_NAMES[1]} on id = student_id ORDER BY average DESC;
        """

        pipeline = PipelineBuilder(
            TEST_CLIENT, name="notebook", sql=sql
        ).create_or_replace()

        df_students = pd.read_csv("tests/students.csv")
        df_grades = pd.read_csv("tests/grades.csv")

        pipeline.start()
        out = pipeline.listen(view_name)

        pipeline.input_pandas(TBL_NAMES[0], df_students)
        pipeline.input_pandas(TBL_NAMES[1], df_grades)

        pipeline.wait_for_completion(True)
        df = out.to_pandas()

        pipeline.delete(True)
        assert df.shape[0] == 100

    def test_two_pipelines(self):
        # https://github.com/feldera/feldera/issues/1770

        TBL_NAMES = ["students", "grades"]
        VIEW_NAMES = [n + "_view" for n in TBL_NAMES]

        sql1 = f"""
        CREATE TABLE {TBL_NAMES[0]} (
            name STRING,
            id INT
        );

        CREATE VIEW {VIEW_NAMES[0]} AS SELECT * FROM {TBL_NAMES[0]};
        """

        sql2 = f"""
        CREATE TABLE {TBL_NAMES[1]} (
            student_id INT,
            science INT,
            maths INT,
            art INT
        );

        CREATE VIEW {VIEW_NAMES[1]} AS SELECT * FROM {TBL_NAMES[1]};
        """

        pipeline1 = PipelineBuilder(
            TEST_CLIENT, name="p1", sql=sql1
        ).create_or_replace()
        pipeline2 = PipelineBuilder(
            TEST_CLIENT, name="p2", sql=sql2
        ).create_or_replace()

        df_students = pd.read_csv("tests/students.csv")
        df_grades = pd.read_csv("tests/grades.csv")

        out1 = pipeline1.listen(VIEW_NAMES[0])
        out2 = pipeline2.listen(VIEW_NAMES[1])

        pipeline1.start()
        pipeline2.start()

        pipeline1.input_pandas(TBL_NAMES[0], df_students)
        pipeline2.input_pandas(TBL_NAMES[1], df_grades)

        pipeline1.wait_for_completion(True)
        pipeline2.wait_for_completion(True)

        df1 = out1.to_pandas()
        df2 = out2.to_pandas()

        assert df1.columns.tolist() not in df2.columns.tolist()

        pipeline1.delete(True)
        pipeline2.delete(True)

    def test_foreach_chunk(self):
        def callback(df: pd.DataFrame, seq_no: int):
            print(f"\nSeq No: {seq_no}, DF size: {df.shape[0]}\n")

        TBL_NAMES = ["students", "grades"]
        view_name = "average_scores"

        sql = f"""
        CREATE TABLE {TBL_NAMES[0]} (
            name STRING,
            id INT
        );

        CREATE TABLE {TBL_NAMES[1]} (
            student_id INT,
            science INT,
            maths INT,
            art INT
        );

        CREATE VIEW {view_name} AS SELECT name, ((science + maths + art) / 3) as average FROM {TBL_NAMES[0]} JOIN {TBL_NAMES[1]} on id = student_id ORDER BY average DESC;
        """

        pipeline = PipelineBuilder(
            TEST_CLIENT, name="foreach_chunk", sql=sql
        ).create_or_replace()

        df_students = pd.read_csv("tests/students.csv")
        df_grades = pd.read_csv("tests/grades.csv")

        pipeline.foreach_chunk(view_name, callback)
        pipeline.start()

        pipeline.input_pandas(TBL_NAMES[0], df_students)
        pipeline.input_pandas(TBL_NAMES[1], df_grades)

        pipeline.wait_for_completion(True)
        pipeline.delete(True)

    def test_df_without_columns(self):
        TBL_NAME = "student"

        sql = f"""
        CREATE TABLE {TBL_NAME} (
            id INT,
            name STRING
        );

        CREATE VIEW s AS SELECT * FROM {TBL_NAME};
        """

        pipeline = PipelineBuilder(
            TEST_CLIENT, name="df_without_columns", sql=sql
        ).create_or_replace()

        df = pd.DataFrame([(1, "a"), (2, "b"), (3, "c")])

        pipeline.start()

        with self.assertRaises(ValueError):
            pipeline.input_pandas(TBL_NAME, df)

        pipeline.stop(force=True)
        pipeline.delete(True)

    def test_sql_error(self):
        pipeline_name = "sql_error"

        sql = """
CREATE TABLE student(
    id INT,
    name STRING
);

CREATE VIEW s AS SELECT * FROM blah;
        """

        expected = f"""
Pipeline {pipeline_name} failed to compile:
Error in SQL statement
Object 'blah' not found
Code snippet:
    6|CREATE VIEW s AS SELECT * FROM blah;
                                     ^^^^""".strip()

        with self.assertRaises(Exception) as err:
            PipelineBuilder(
                TEST_CLIENT, name=pipeline_name, sql=sql
            ).create_or_replace()

        got_err: str = err.exception.args[0].strip()

        assert expected == got_err

        pipeline = Pipeline.get("sql_error", TEST_CLIENT)
        pipeline.delete(True)

    def test_kafka(self):
        import json

        in_ci = os.environ.get("IN_CI")

        if in_ci == "1":
            # if running in CI, skip the test
            return

        print("(Re-)creating topics...")
        admin_client = KafkaAdminClient(
            bootstrap_servers=KAFKA_SERVER, client_id="test_client"
        )

        INPUT_TOPIC = "simple_count_input"
        OUTPUT_TOPIC = "simple_count_output"

        existing_topics = set(admin_client.list_topics())
        if INPUT_TOPIC in existing_topics:
            admin_client.delete_topics([INPUT_TOPIC])
        if OUTPUT_TOPIC in existing_topics:
            admin_client.delete_topics([OUTPUT_TOPIC])
        admin_client.create_topics(
            [
                NewTopic(INPUT_TOPIC, num_partitions=1, replication_factor=1),
                NewTopic(OUTPUT_TOPIC, num_partitions=1, replication_factor=1),
            ]
        )
        print("Topics ready")

        # Produce rows into the input topic
        print("Producing rows into input topic...")
        num_rows = 1000
        producer = KafkaProducer(
            bootstrap_servers=KAFKA_SERVER,
            client_id="test_client",
            value_serializer=lambda v: json.dumps(v).encode("utf-8"),
        )
        for i in range(num_rows):
            producer.send("simple_count_input", value={"insert": {"id": i}})
        print("Input topic contains data")

        TABLE_NAME = "example"
        VIEW_NAME = "example_count"

        sql = f"""
        CREATE TABLE {TABLE_NAME} (id INT NOT NULL PRIMARY KEY)
        WITH (
            'connectors' = '[
                {{
                    "name": "kafka-2",
                    "transport": {{
                        "name": "kafka_input",
                        "config": {{
                            "bootstrap.servers": "{PIPELINE_TO_KAFKA_SERVER}",
                            "topics": ["{INPUT_TOPIC}"],
                            "auto.offset.reset": "earliest"
                        }}
                    }},
                    "format": {{
                        "name": "json",
                        "config": {{
                            "update_format": "insert_delete",
                            "array": false
                        }}
                    }}
                }}
            ]'
        );

        CREATE VIEW {VIEW_NAME}
        WITH (
            'connectors' = '[
                {{
                    "name": "kafka-3",
                    "transport": {{
                        "name": "kafka_output",
                        "config": {{
                            "bootstrap.servers": "{PIPELINE_TO_KAFKA_SERVER}",
                            "topic": "{OUTPUT_TOPIC}",
                            "auto.offset.reset": "earliest"
                        }}
                    }},
                    "format": {{
                        "name": "json",
                        "config": {{
                            "update_format": "insert_delete",
                            "array": false
                        }}
                    }}
                }}
            ]'
        )
        AS SELECT COUNT(*) as num_rows FROM {TABLE_NAME};
        """

        pipeline = PipelineBuilder(
            TEST_CLIENT, name="kafka_test", sql=sql
        ).create_or_replace()

        out = pipeline.listen(VIEW_NAME)
        pipeline.start()
        pipeline.wait_for_idle()
        pipeline.stop(force=True)
        df = out.to_pandas()
        assert df.shape[0] != 0

        pipeline.delete(True)

    def test_http_get(self):
        sql = """
        CREATE TABLE items (
            id INT,
            name STRING
        ) WITH (
            'connectors' = '[
                {
                    "name": "url_conn",
                    "transport": {
                        "name": "url_input",
                        "config": {
                            "path": "https://feldera-basics-tutorial.s3.amazonaws.com/part.json"
                        }
                    },
                    "format": {
                        "name": "json",
                        "config": {
                            "update_format": "insert_delete",
                            "array": false
                        }
                    }
                }
            ]'
        );

        CREATE MATERIALIZED VIEW s AS SELECT * FROM items;
        """

        pipeline = PipelineBuilder(
            TEST_CLIENT, name="test_http_get", sql=sql
        ).create_or_replace()

        out = pipeline.listen("s")

        pipeline.start()
        pipeline.wait_for_completion(True)

        df = out.to_pandas()

        assert df.shape[0] == 3

        pipeline.delete(True)

    def test_avro_format(self):
        import json

        TOPIC = "test_avro_format"

        in_ci = os.environ.get("IN_CI")

        if in_ci == "1":
            # if running in CI, skip the test
            return

        admin_client = KafkaAdminClient(
            bootstrap_servers=KAFKA_SERVER, client_id="test_client"
        )
        existing_topics = set(admin_client.list_topics())
        if TOPIC in existing_topics:
            admin_client.delete_topics([TOPIC])

        df = pd.DataFrame({"id": [1, 2, 3], "name": ["a", "b", "c"]})

        sql = f"""
        CREATE TABLE items (
            id INT,
            name STRING
        );

        CREATE VIEW s WITH (
           'connectors'  = '[
                {{
                    "name": "kafka-1",
                    "transport": {{
                        "name": "kafka_output",
                        "config": {{
                            "bootstrap.servers": "{PIPELINE_TO_KAFKA_SERVER}",
                            "topic": "{TOPIC}"
                        }}
                    }},
                    "format": {{
                        "name": "avro",
                        "config": {{
                            "schema": {
            json.dumps(
                json.dumps(
                    {
                        "type": "record",
                        "name": "items",
                        "fields": [
                            {"name": "id", "type": ["null", "int"]},
                            {"name": "name", "type": ["null", "string"]},
                        ],
                    }
                )
            )
        }
                        }}
                    }}
                }}
           ]'
        )
        AS SELECT * FROM items;
        """

        pipeline = PipelineBuilder(
            TEST_CLIENT, name="test_avro_format", sql=sql
        ).create_or_replace()

        pipeline.start()
        pipeline.input_pandas("items", df)
        pipeline.wait_for_completion(True)

        consumer = KafkaConsumer(
            TOPIC,
            bootstrap_servers=KAFKA_SERVER,
            auto_offset_reset="earliest",
        )

        msg = next(consumer)
        assert msg.value is not None

        pipeline.delete(True)

    def test_pipeline_resource_config(self):
        from feldera.runtime_config import Resources, RuntimeConfig

        config = {
            "cpu_cores_max": 3,
            "cpu_cores_min": 2,
            "memory_mb_max": 500,
            "memory_mb_min": 300,
            "storage_mb_max": None,
            "storage_class": None,
        }

        resources = Resources(config)
        name = "test_pipeline_resource_config"

        sql = """
        CREATE TABLE items (
            id INT,
            name STRING
        ) WITH (
            'connectors' = '[
                {
                    "name": "url_conn",
                    "transport": {
                        "name": "url_input",
                        "config": {
                            "path": "https://feldera-basics-tutorial.s3.amazonaws.com/part.json"
                        }
                    },
                    "format": {
                        "name": "json",
                        "config": {
                            "update_format": "insert_delete",
                            "array": false
                        }
                    }
                }
            ]'
        );

        CREATE VIEW s AS SELECT * FROM items;
        """

        pipeline = PipelineBuilder(
            TEST_CLIENT,
            name=name,
            sql=sql,
            runtime_config=RuntimeConfig(
                resources=resources, storage=False, workers=10
            ),
        ).create_or_replace()

        out = pipeline.listen("s")

        pipeline.start()
        pipeline.wait_for_completion(True)

        df = out.to_pandas()

        assert df.shape[0] == 3

        assert TEST_CLIENT.get_pipeline(name).runtime_config["resources"] == config

        pipeline.delete(True)

    def test_timestamp_pandas(self):
        TBL_NAME = "items"
        VIEW_NAME = "s"

        sql = f"""
        CREATE TABLE {TBL_NAME} (
            id INT,
            name STRING,
            birthdate TIMESTAMP
        );

        CREATE VIEW {VIEW_NAME} AS SELECT * FROM {TBL_NAME};
        """

        pipeline = PipelineBuilder(
            TEST_CLIENT, name="test_timestamp_pandas", sql=sql
        ).create_or_replace()

        df = pd.DataFrame(
            {
                "id": [1, 2, 3],
                "name": ["a", "b", "c"],
                "birthdate": [
                    pd.Timestamp.now(),
                    pd.Timestamp.now(),
                    pd.Timestamp.now(),
                ],
            }
        )

        out = pipeline.listen(VIEW_NAME)

        pipeline.start()
        pipeline.input_pandas(TBL_NAME, df)
        pipeline.wait_for_completion(True)

        df = out.to_pandas()

        assert df.shape[0] == 3

        pipeline.delete(True)

    def test_input_json0(self):
        TBL_NAME = "items"
        VIEW_NAME = "s"

        sql = f"""
        CREATE TABLE {TBL_NAME} (
            id INT,
            name STRING
        );

        CREATE VIEW {VIEW_NAME} AS SELECT * FROM {TBL_NAME};
        """

        pipeline = PipelineBuilder(
            TEST_CLIENT, name="test_input_json", sql=sql
        ).create_or_replace()

        data = {"insert": {"id": 1, "name": "a"}}

        out = pipeline.listen(VIEW_NAME)

        pipeline.start()
        pipeline.input_json(TBL_NAME, data, update_format="insert_delete")
        pipeline.wait_for_completion(True)

        out_data = out.to_dict()

        data["insert"].update({"insert_delete": 1})
        assert out_data == [data["insert"]]

        pipeline.delete(True)

    def test_input_json1(self):
        TBL_NAME = "items"
        VIEW_NAME = "s"

        sql = f"""
        CREATE TABLE {TBL_NAME} (
            id INT,
            name STRING
        );

        CREATE VIEW {VIEW_NAME} AS SELECT * FROM {TBL_NAME};
        """

        pipeline = PipelineBuilder(
            TEST_CLIENT, name="test_input_json", sql=sql
        ).create_or_replace()

        data = [{"id": 1, "name": "a"}, {"id": 2, "name": "b"}]

        out = pipeline.listen(VIEW_NAME)

        pipeline.start()
        pipeline.input_json(TBL_NAME, data)
        pipeline.wait_for_completion(True)

        out_data = out.to_dict()

        for datum in data:
            datum.update({"insert_delete": 1})

        assert out_data == data

        pipeline.delete(True)

    def test_suspend(self):
        if not TEST_CLIENT.get_config().edition.is_enterprise():
            return

        TBL_NAME = "items"
        VIEW_NAME = "s"

        sql = f"""
        CREATE TABLE {TBL_NAME} (
            id INT,
            name STRING
        );

        CREATE VIEW {VIEW_NAME} AS SELECT * FROM {TBL_NAME};
        """

        pipeline = PipelineBuilder(
            TEST_CLIENT, name="test_suspend", sql=sql
        ).create_or_replace()

        data = {"insert": {"id": 1, "name": "a"}}

        out = pipeline.listen(VIEW_NAME)

        pipeline.start()
        pipeline.input_json(TBL_NAME, data, update_format="insert_delete")
        pipeline.wait_for_completion(False)

        pipeline.stop(force=False)

        out_data = out.to_dict()

        data["insert"].update({"insert_delete": 1})
        assert out_data == [data["insert"]]

        pipeline.stop(force=True)
        pipeline.delete(True)

    def test_issue2142(self):
        sql = """
        CREATE TABLE t0 (c1 INT);
        CREATE VIEW v0 AS SELECT * FROM t0;
        """

        pipeline = PipelineBuilder(
            TEST_CLIENT, name="test_issue2142", sql=sql
        ).create_or_replace()

        data = [{"c1": None}, {"c1": 1}]

        out = pipeline.listen("v0")

        pipeline.start()
        pipeline.input_json("t0", data=data)
        pipeline.wait_for_completion(True)

        out_data = out.to_dict()

        for datum in data:
            datum.update({"insert_delete": 1})

        assert out_data == data

        pipeline.delete(True)

    def test_pandas_binary(self):
        sql = """
        CREATE TABLE t0 (c1 VARBINARY);
        CREATE VIEW v0 AS SELECT SUBSTRING(c1 FROM 2) as c1 FROM t0;
        """

        data = [{"c1": [12, 34, 56]}]
        expected_data = [{"c1": [34, 56], "insert_delete": 1}]

        pipeline = PipelineBuilder(
            TEST_CLIENT, name="test_pandas_binary", sql=sql
        ).create_or_replace()
        out = pipeline.listen("v0")

        pipeline.start()
        pipeline.input_json("t0", data=data)
        pipeline.wait_for_completion(True)

        got = out.to_dict()

        assert expected_data == got
        pipeline.delete(True)

    def test_pandas_decimal(self):
        from decimal import Decimal

        sql = """
        CREATE TABLE t0 (c1 DECIMAL(5, 2));
        CREATE VIEW v0 AS SELECT c1 + 2.75::DECIMAL(5, 2) as c1 FROM t0;
        """

        data = [{"c1": 2.25}]
        expected = [{"c1": Decimal("5.00"), "insert_delete": 1}]

        pipeline = PipelineBuilder(
            TEST_CLIENT, name="test_pandas_decimal", sql=sql
        ).create_or_replace()
        out = pipeline.listen("v0")

        pipeline.start()
        pipeline.input_json("t0", data=data)
        pipeline.wait_for_completion(True)

        got = out.to_dict()

        assert expected == got
        pipeline.delete(True)

    def test_pandas_array(self):
        sql = """
        CREATE TABLE t0 (c1 INT ARRAY);
        CREATE VIEW v0 AS SELECT c1 FROM t0;
        """

        data = [{"c1": [1, 2, 3]}]

        pipeline = PipelineBuilder(
            TEST_CLIENT, name="test_pandas_array", sql=sql
        ).create_or_replace()
        out = pipeline.listen("v0")
        pipeline.start()
        pipeline.input_json("t0", data=data)
        pipeline.wait_for_completion(True)

        got = out.to_dict()

        for datum in data:
            datum.update({"insert_delete": 1})

        assert got == data
        pipeline.delete(True)

    def test_pandas_struct(self):
        sql = """
        CREATE TYPE s AS (
            f1 INT,
            f2 STRING
        );
        CREATE TABLE t0 (c1 s);
        CREATE VIEW v0 AS SELECT c1 FROM t0;
        """

        data = [{"c1": {"f1": 1, "f2": "a"}}]
        pipeline = PipelineBuilder(
            TEST_CLIENT, name="test_pandas_struct", sql=sql
        ).create_or_replace()
        out = pipeline.listen("v0")
        pipeline.start()
        pipeline.input_json("t0", data)
        pipeline.wait_for_completion(True)
        got = out.to_dict()

        for datum in data:
            datum.update({"insert_delete": 1})

        assert data == got
        pipeline.delete(True)

    def test_pandas_date_time_timestamp(self):
        from pandas import Timestamp, Timedelta

        sql = """
        CREATE TABLE t0 (c1 DATE, c2 TIME, c3 TIMESTAMP);
        CREATE VIEW v0 AS SELECT c1, c2, c3 FROM t0;
        """

        data = [{"c1": "2022-01-01", "c2": "12:00:00", "c3": "2022-01-01 12:00:00"}]
        expected = [
            {
                "c1": Timestamp("2022-01-01 00:00:00"),
                "c2": Timedelta("0 days 12:00:00"),
                "c3": Timestamp("2022-01-01 12:00:00"),
                "insert_delete": 1,
            }
        ]

        pipeline = PipelineBuilder(
            TEST_CLIENT, name="test_pandas_date_time_timestamp", sql=sql
        ).create_or_replace()
        out = pipeline.listen("v0")
        pipeline.start()
        pipeline.input_json("t0", data)
        pipeline.wait_for_completion(True)
        got = out.to_dict()

        assert expected == got
        pipeline.delete(True)

    def test_pandas_simple(self):
        sql = """
        CREATE TABLE t0 (c0 BOOLEAN, c1 TINYINT, c2 SMALLINT, c3 INT, c4 BIGINT, c5 REAL, c6 DOUBLE, c7 VARCHAR, c8 CHAR);
        CREATE VIEW v0 AS SELECT * FROM t0;
        """

        data = [
            {
                "c0": True,
                "c1": 1,
                "c2": 2,
                "c3": 3,
                "c4": 4,
                "c5": 5.0,
                "c6": 6.0,
                "c7": "seven",
                "c8": "c",
            }
        ]

        pipeline = PipelineBuilder(
            TEST_CLIENT, name="test_pandas_simple", sql=sql
        ).create_or_replace()
        out = pipeline.listen("v0")
        pipeline.start()
        pipeline.input_json("t0", data)
        pipeline.wait_for_completion(True)
        got = out.to_dict()

        for datum in data:
            datum.update({"insert_delete": 1})

        assert data == got
        pipeline.delete(True)

    def test_pandas_map(self):
        sql = """
        CREATE TABLE t0 (c1 MAP<STRING, INT>);
        CREATE VIEW v0 AS SELECT c1 FROM t0;
        """

        data = [{"c1": {"a": 1, "b": 2}}]
        expected = [{"c1": {"a": 1, "b": 2}, "insert_delete": 1}]

        pipeline = PipelineBuilder(
            TEST_CLIENT, name="test_pandas_map", sql=sql
        ).create_or_replace()
        out = pipeline.listen("v0")
        pipeline.start()
        pipeline.input_json("t0", data)
        pipeline.wait_for_completion(True)
        pipeline.stop(force=True)

        got = out.to_dict()
        assert expected == got

        pipeline.start()
        out = pipeline.listen("v0")
        pipeline.input_json("t0", {"c1": {"a": 1, "b": 2}})
        pipeline.wait_for_completion(True)
        pipeline.stop(force=True)

        got = out.to_dict()
        assert expected == got

        pipeline.delete(True)

    def test_failed_pipeline_stop(self):
        sql = """
            CREATE TABLE t0 (c1 TINYINT);
            CREATE VIEW v0 AS SELECT c1 + 127::TINYINT FROM t0;"""

        pipeline = PipelineBuilder(
            TEST_CLIENT, name="test_failed_pipeline_stop", sql=sql
        ).create_or_replace()
        pipeline.start()
        data = [{"c1": 127}]
        pipeline.input_json("t0", data)

        while True:
            status = pipeline.status()
            expected = PipelineStatus.STOPPED
            if status == expected and len(pipeline.deployment_error()) > 0:
                break
            time.sleep(1)

        pipeline.stop(force=True)
        pipeline.delete(True)

    def test_adhoc_execute(self):
        sql = """
        CREATE TABLE t0 (c1 TINYINT) with ('materialized' = 'true');
        """

        pipeline = PipelineBuilder(
            TEST_CLIENT, name="test_adhoc_execute", sql=sql
        ).create_or_replace()
        pipeline.start()
        pipeline.wait_for_completion()
        pipeline.execute("INSERT INTO t0 VALUES (1), (2);")
        resp = pipeline.query("SELECT * FROM t0;")

        got = list(resp)
        expected = [{"c1": 1}, {"c1": 2}]

        pipeline.stop(force=True)

        self.assertCountEqual(got, expected)

    def test_initialization_error(self):
        sql = """
        CREATE TABLE t0 (
            c0 INT NOT NULL
        ) with (
          'connectors' = '[{
            "transport": {
              "name": "datagen",
              "config": {
                "plan": [{
                    "fields": {
                        "c1": { "strategy": "uniform", "range": [100, 10000] }
                    }
                }]
              }
            }
          }]'
        );
        """

        pipeline = PipelineBuilder(
            TEST_CLIENT, name="test_initialization_error", sql=sql
        ).create_or_replace()
        with self.assertRaises(RuntimeError) as err:
            pipeline.start()

        pipeline.stop(force=True)
        pipeline.delete(True)

        got_err: str = err.exception.args[0].strip()
        assert "Unable to START the pipeline" in got_err

    def test_connector_orchestration(self):
        sql = """
        CREATE TABLE numbers (
          num INT
        ) WITH (
            'connectors' = '[
                {
                    "name": "c1",
                    "paused": true,
                    "transport": {
                        "name": "datagen",
                        "config": {"plan": [{ "rate": 1, "fields": { "num": { "range": [0, 10], "strategy": "uniform" } } }]}
                    }
                }
            ]'
        );
        """

        name = "test_connector_orchestration"

        pipeline = PipelineBuilder(TEST_CLIENT, name, sql=sql).create_or_replace()
        pipeline.start()

        pipeline.resume_connector("numbers", "c1")
        stats = TEST_CLIENT.get_pipeline_stats(name)
        c1_status = next(
            item["paused"]
            for item in stats["inputs"]
            if item["endpoint_name"] == "numbers.c1"
        )
        assert not c1_status

        pipeline.pause_connector("numbers", "c1")
        stats = TEST_CLIENT.get_pipeline_stats(name)
        c2_status = next(
            item["paused"]
            for item in stats["inputs"]
            if item["endpoint_name"] == "numbers.c1"
        )
        assert c2_status

        pipeline.stop(force=True)
        pipeline.delete(True)

    def test_uuid(self):
        name = "test_uuid"
        sql = """
        CREATE TABLE t0(c0 UUID);
        CREATE MATERIALIZED VIEW v0 AS SELECT c0 FROM t0;
        """

        pipeline = PipelineBuilder(TEST_CLIENT, name, sql=sql).create_or_replace()
        out = pipeline.listen("v0")
        pipeline.start()

        data = [{"c0": uuid.uuid4()}]
        pipeline.input_json("t0", data)

        for datum in data:
            datum.update({"insert_delete": 1})

        pipeline.wait_for_completion(True)

        got = out.to_dict()
        pipeline.stop(force=True)

        assert got == data

    def test_issue3754(self):
        name = "test_issue3754"
        sql = """
CREATE TABLE map_tbl(m_var MAP<VARCHAR, VARCHAR>);
CREATE MATERIALIZED VIEW v AS SELECT * FROM map_tbl;
        """

        pipeline = PipelineBuilder(TEST_CLIENT, name, sql=sql).create_or_replace()
        pipeline.start()

        with self.assertRaises(ValueError):
            data = [{"insert": {"m_var": {None: 1}}}]
            TEST_CLIENT.push_to_pipeline(name, "map_tbl", "insert_delete", data)

        with self.assertRaises(ValueError):
            data = [{"delete": {"m_var": {None: 1}}}]
            TEST_CLIENT.push_to_pipeline(name, "map_tbl", "insert_delete", data)

        with self.assertRaises(ValueError):
            data = {"insert": {"m_var": {None: 1}}}
            TEST_CLIENT.push_to_pipeline(name, "map_tbl", "insert_delete", data)

        with self.assertRaises(ValueError):
            data = {"delete": {"m_var": {None: 1}}}
            TEST_CLIENT.push_to_pipeline(name, "map_tbl", "insert_delete", data)

        with self.assertRaises(ValueError):
            data = [{"m_var": {None: 1}}]
            TEST_CLIENT.push_to_pipeline(name, "map_tbl", "raw", data)

        with self.assertRaises(ValueError):
            data = {"m_var": {None: 1}}
            TEST_CLIENT.push_to_pipeline(name, "map_tbl", "raw", data)

        with self.assertRaises(ValueError):
            data = {"m_var": {None: 1}}
            pipeline.input_json("map_tbl", data)

        pipeline.stop(force=True)
        pipeline.delete(True)

    def test_program_error0(self):
        sql = "create taabl;"
        name = "test_program_error0"

        try:
            _ = PipelineBuilder(TEST_CLIENT, name, sql).create_or_replace()
        except Exception:
            pass

        pipeline = Pipeline.get(name, TEST_CLIENT)
        err = pipeline.program_error()

        assert err["sql_compilation"] != 0

        pipeline.stop(force=True)
        pipeline.delete(True)

    def test_program_error1(self):
        sql = ""
        name = "test_program_error1"

        _ = PipelineBuilder(TEST_CLIENT, name, sql).create_or_replace()

        pipeline = Pipeline.get(name, TEST_CLIENT)
        err = pipeline.program_error()

        assert err["sql_compilation"]["exit_code"] == 0
        assert err["rust_compilation"]["exit_code"] == 0

        pipeline.stop(force=True)
        pipeline.delete(True)

    def test_errors0(self):
        sql = "SELECT invalid"
        name = "test_errors0"

        try:
            _ = PipelineBuilder(TEST_CLIENT, name, sql).create_or_replace()
        except Exception:
            pass

        pipeline = Pipeline.get(name, TEST_CLIENT)

        assert pipeline.errors()[0]["sql_compilation"]["exit_code"] != 0


if __name__ == "__main__":
    unittest.main()



from enum import Enum
from typing import Optional


class CompilationProfile(Enum):
    """
    The compilation profile to use when compiling the program.
    """

    SERVER_DEFAULT = None
    """
    The compiler server default compilation profile.
    """

    DEV = "dev"
    """
    The development compilation profile.
    """

    UNOPTIMIZED = "unoptimized"
    """
    The unoptimized compilation profile.
    """

    OPTIMIZED = "optimized"
    """
    The optimized compilation profile, the default for this API.
    """


class BuildMode(Enum):
    CREATE = 1
    GET = 2
    GET_OR_CREATE = 3


class PipelineStatus(Enum):
    """
      Represents the state that this pipeline is currently in.

      .. code-block:: text

                  Stopped ◄─────────── Stopping ◄───── All states can transition
                     │                    ▲            to Stopping by either:
    /start or /pause │                    │            (1) user calling /stop?force=true, or;
                     ▼                    │            (2) pipeline encountering a fatal
              ⌛Provisioning          Suspending            resource or runtime error,
                     │                    ▲                having the system call /stop?force=true
                     ▼                    │ /stop          effectively
              ⌛Initializing ─────────────┤  ?force=false
                     │                    │
           ┌─────────┼────────────────────┴─────┐
           │         ▼                          │
           │       Paused  ◄──────► Unavailable │
           │        │   ▲                ▲      │
           │ /start │   │  /pause        │      │
           │        ▼   │                │      │
           │       Running ◄─────────────┘      │
           └────────────────────────────────────┘
    """

    NOT_FOUND = 0
    """
    The pipeline has not been created yet.
    """

    STOPPED = 1
    """
    The pipeline has not (yet) been started or has been stopped either
    manually by the user or automatically by the system due to a
    resource or runtime error.

    The pipeline remains in this state until:

        1. The user starts it via `/start` or `/pause`, transitioning to `PROVISIONING`.
        2. Early start fails (e.g., compilation failure), transitioning to `STOPPING`.
    """

    PROVISIONING = 2
    """
    Compute (and optionally storage) resources needed for running the pipeline
    are being provisioned.

    The pipeline remains in this state until:

        1. Resources are provisioned successfully, transitioning to `INITIALIZING`.
        2. Provisioning fails or times out, transitioning to `STOPPING`.
        3. The user cancels the pipeline via `/stop`, transitioning to `STOPPING`.
    """

    INITIALIZING = 3
    """
    The pipeline is initializing its internal state and connectors.

    The pipeline remains in this state until:

        1. Initialization succeeds, transitioning to `PAUSED`.
        2. Initialization fails or times out, transitioning to `STOPPING`.
        3. The user suspends the pipeline via `/suspend`, transitioning to `SUSPENDING`.
        4. The user stops the pipeline via `/stop`, transitioning to `STOPPING`.
    """

    PAUSED = 4
    """
    The pipeline is initialized but data processing is paused.

    The pipeline remains in this state until:

        1. The user starts it via `/start`, transitioning to `RUNNING`.
        2. A runtime error occurs, transitioning to `STOPPING`.
        3. The user suspends it via `/suspend`, transitioning to `SUSPENDING`.
        4. The user stops it via `/stop`, transitioning to `STOPPING`.
    """

    RUNNING = 5
    """
    The pipeline is processing data.

    The pipeline remains in this state until:

        1. The user pauses it via `/pause`, transitioning to `PAUSED`.
        2. A runtime error occurs, transitioning to `STOPPING`.
        3. The user suspends it via `/suspend`, transitioning to `SUSPENDING`.
        4. The user stops it via `/stop`, transitioning to `STOPPING`.
    """

    UNAVAILABLE = 6
    """
    The pipeline was initialized at least once but is currently unreachable
    or not ready.

    The pipeline remains in this state until:

        1. A successful status check transitions it back to `PAUSED` or `RUNNING`.
        2. A runtime error occurs, transitioning to `STOPPING`.
        3. The user suspends it via `/suspend`, transitioning to `SUSPENDING`.
        4. The user stops it via `/stop`, transitioning to `STOPPING`.

    Note: While in this state, `/start` or `/pause` express desired state but
    are only applied once the pipeline becomes reachable.
    """

    SUSPENDING = 7
    """
    The pipeline is being suspended to storage.

    The pipeline remains in this state until:

        1. Suspension succeeds, transitioning to `STOPPING`.
        2. A runtime error occurs, transitioning to `STOPPING`.
    """

    STOPPING = 8
    """
    The pipeline's compute resources are being scaled down to zero.

    The pipeline remains in this state until deallocation completes,
    transitioning to `STOPPED`.
    """

    @staticmethod
    def from_str(value):
        for member in PipelineStatus:
            if member.name.lower() == value.lower():
                return member
        raise ValueError(f"Unknown value '{value}' for enum {PipelineStatus.__name__}")

    def __eq__(self, other):
        return self.value == other.value


class ProgramStatus(Enum):
    Pending = 1
    CompilingSql = 2
    SqlCompiled = 3
    CompilingRust = 4
    Success = 5
    SqlError = 6
    RustError = 7
    SystemError = 8

    def __init__(self, value):
        self.error: Optional[dict] = None
        self._value_ = value

    @staticmethod
    def from_value(value):
        error = None
        if isinstance(value, dict):
            error = value
            value = list(value.keys())[0]

        for member in ProgramStatus:
            if member.name.lower() == value.lower():
                member.error = error
                return member
        raise ValueError(f"Unknown value '{value}' for enum {ProgramStatus.__name__}")

    def __eq__(self, other):
        return self.value == other.value

    def __str__(self):
        return self.name + (f": ({self.error})" if self.error else "")

    def get_error(self) -> Optional[dict]:
        """
        Returns the compilation error, if any.
        """

        return self.error


class CheckpointStatus(Enum):
    Success = 1
    Failure = 2
    InProgress = 3
    Unknown = 4

    def __init__(self, value):
        self.error: Optional[str] = None
        self._value_ = value

    def __eq__(self, other):
        return self.value == other.value

    def get_error(self) -> Optional[str]:
        """
        Returns the error, if any.
        """

        return self.error


class StorageStatus(Enum):
    """
    Represents the current storage usage status of the pipeline.
    """

    CLEARED = 0
    """
    The pipeline has not been started before, or the user has cleared storage.

    In this state, the pipeline has no storage resources bound to it.
    """

    INUSE = 1
    """
    The pipeline was (attempted to be) started before, transitioning from `STOPPED`
    to `PROVISIONING`, which caused the storage status to become `INUSE`.

    Being in the `INUSE` state restricts certain edits while the pipeline is `STOPPED`.

    The pipeline remains in this state until the user invokes `/clear`, transitioning
    it to `CLEARING`.
    """

    CLEARING = 2
    """
    The pipeline is in the process of becoming unbound from its storage resources.

    If storage resources are configured to be deleted upon clearing, their deletion
    occurs before transitioning to `CLEARED`. Otherwise, no actual work is required,
    and the transition happens immediately.

    If storage is not deleted during clearing, the responsibility to manage or delete
    those resources lies with the user.
    """

    @staticmethod
    def from_str(value):
        for member in StorageStatus:
            if member.name.lower() == value.lower():
                return member
        raise ValueError(f"Unknown value '{value}' for enum {StorageStatus.__name__}")

    def __eq__(self, other):
        return self.value == other.value



import os
import pathlib
import unittest
import uuid
import threading

from tests import TEST_CLIENT

from feldera.rest.pipeline import Pipeline


class TestPipeline(unittest.TestCase):
    result = None

    def test_delete_all_pipelines(self):
        pipelines = TEST_CLIENT.pipelines()
        for pipeline in pipelines:
            TEST_CLIENT.stop_pipeline(pipeline.name, force=True)
            if pipeline.storage_status != "Cleared":
                TEST_CLIENT.clear_storage(pipeline.name)
            TEST_CLIENT.delete_pipeline(pipeline.name)

    def test_create_pipeline(
        self, name: str = "blah", delete=False, runtime_config: dict = {}
    ):
        self.test_delete_all_pipelines()
        sql = """
        CREATE TABLE tbl(id INT) WITH ('append_only' = 'true');
        CREATE VIEW V AS SELECT * FROM tbl;
        """
        pipeline = Pipeline(name, sql, "", "", {}, runtime_config)
        pipeline = TEST_CLIENT.create_pipeline(pipeline)

        if pipeline.storage_status != "Cleared":
            TEST_CLIENT.clear_storage(pipeline.name)

        if delete:
            if pipeline.storage_status != "Cleared":
                TEST_CLIENT.clear_storage(pipeline.name)
            TEST_CLIENT.delete_pipeline(pipeline.name)

    def test_delete_pipeline(self):
        name = str(uuid.uuid4())
        self.test_create_pipeline(name, False)

        p = TEST_CLIENT.get_pipeline(name)
        if p.storage_status != "Cleared":
            TEST_CLIENT.clear_storage(p.name)

        TEST_CLIENT.delete_pipeline(name)

    def __test_push_to_pipeline(self, data, format, array):
        name = str(uuid.uuid4())
        self.test_create_pipeline(name, False)

        TEST_CLIENT.start_pipeline(name)

        TEST_CLIENT.push_to_pipeline(
            pipeline_name=name,
            table_name="tbl",
            format=format,
            array=array,
            data=data,
        )

        TEST_CLIENT.pause_pipeline(name)
        TEST_CLIENT.stop_pipeline(name, force=True)

        TEST_CLIENT.clear_storage(name)
        TEST_CLIENT.delete_pipeline(name)

    def test_push_to_pipeline_json(self):
        data = [
            {"id": 1},
            {"id": 2},
            {"id": 3},
        ]
        self.__test_push_to_pipeline(data, format="json", array=True)

    def test_push_to_pipeline_csv0(self):
        data = "1\n2\n"
        self.__test_push_to_pipeline(data, format="csv", array=False)

    def test_list_pipelines(self):
        name = str(uuid.uuid4())
        self.test_create_pipeline(name, False)
        pipelines = TEST_CLIENT.pipelines()
        assert len(pipelines) > 0
        assert name in [p.name for p in pipelines]

        for p in pipelines:
            if p.storage_status != "Cleared":
                TEST_CLIENT.clear_storage(name)
            TEST_CLIENT.delete_pipeline(name)

    def test_get_pipeline(self):
        name = str(uuid.uuid4())
        self.test_create_pipeline(name, False)
        p = TEST_CLIENT.get_pipeline(name)
        assert name == p.name

        if p.storage_status != "Cleared":
            TEST_CLIENT.clear_storage(name)
        TEST_CLIENT.delete_pipeline(name)

    def test_get_pipeline_config(self):
        name = str(uuid.uuid4())
        self.test_create_pipeline(name, False, {"workers": 2, "storage": False})
        config = TEST_CLIENT.get_runtime_config(name)

        assert config is not None
        assert config.get("workers") == 2
        assert not config.get("storage")

        TEST_CLIENT.delete_pipeline(name)

    def test_get_pipeline_stats(self):
        name = str(uuid.uuid4())
        self.test_create_pipeline(name, delete=False)

        TEST_CLIENT.start_pipeline(name)

        stats = TEST_CLIENT.get_pipeline_stats(name)

        assert stats is not None
        assert stats.get("global_metrics") is not None
        assert stats.get("inputs") is not None
        assert stats.get("outputs") is not None

        TEST_CLIENT.pause_pipeline(name)
        TEST_CLIENT.stop_pipeline(name, force=True)
        TEST_CLIENT.clear_storage(name)
        TEST_CLIENT.delete_pipeline(name)

    def __listener(self, name: str):
        gen_obj = TEST_CLIENT.listen_to_pipeline(
            pipeline_name=name,
            table_name="V",
            format="csv",
        )
        counter = 0
        for chunk in gen_obj:
            counter += 1
            text_data = chunk.get("text_data")
            if text_data:
                assert text_data == "1,1\n2,1\n"
                self.result = True
                break
            if counter > 10:
                self.result = False
                break

    def test_listen_to_pipeline(self):
        data = "1\n2\n"
        name = str(uuid.uuid4())
        self.test_create_pipeline(name, False)

        TEST_CLIENT.pause_pipeline(name)

        t1 = threading.Thread(target=self.__listener, args=(name,))
        t1.start()

        TEST_CLIENT.start_pipeline(name)
        TEST_CLIENT.push_to_pipeline(name, "tbl", "csv", data)

        t1.join()

        assert self.result

        TEST_CLIENT.stop_pipeline(name, force=True)
        TEST_CLIENT.clear_storage(name)
        TEST_CLIENT.delete_pipeline(name)

    def test_adhoc_query_text(self):
        data = "1\n2\n"
        name = str(uuid.uuid4())

        sql = """
        CREATE TABLE tbl(id INT) with ('materialized' = 'true');
        """

        pipeline = Pipeline(name, sql, "", "", {}, {})
        pipeline = TEST_CLIENT.create_pipeline(pipeline)

        TEST_CLIENT.start_pipeline(name)

        TEST_CLIENT.push_to_pipeline(name, "tbl", "csv", data)
        resp = TEST_CLIENT.query_as_text(pipeline.name, "SELECT * FROM tbl")
        expected = [
            """+----+
| id |
+----+
| 1  |
| 2  |
+----+""",
            """+----+
| id |
+----+
| 2  |
| 1  |
+----+""",
        ]

        got = "\n".join(resp)
        assert got in expected

        TEST_CLIENT.stop_pipeline(name, force=True)
        TEST_CLIENT.clear_storage(name)
        TEST_CLIENT.delete_pipeline(name)

    def test_adhoc_query_parquet(self):
        data = "1\n2\n"
        name = str(uuid.uuid4())

        sql = """
        CREATE TABLE tbl(id INT) with ('materialized' = 'true');
        """

        pipeline = Pipeline(name, sql, "", "", {}, {})
        pipeline = TEST_CLIENT.create_pipeline(pipeline)

        TEST_CLIENT.start_pipeline(name)

        TEST_CLIENT.push_to_pipeline(name, "tbl", "csv", data)

        file = name.split("-")[0]
        TEST_CLIENT.query_as_parquet(pipeline.name, "SELECT * FROM tbl", file)
        TEST_CLIENT.stop_pipeline(name, force=True)
        TEST_CLIENT.clear_storage(name)
        TEST_CLIENT.delete_pipeline(name)

        path = pathlib.Path(file + ".parquet")
        assert path.stat().st_size > 0

        os.remove(path)

    def test_adhoc_query_json(self):
        data = "1\n2\n"
        name = str(uuid.uuid4())

        sql = """
        CREATE TABLE tbl(id INT) with ('materialized' = 'true');
        """

        pipeline = Pipeline(name, sql, "", "", {}, {})
        pipeline = TEST_CLIENT.create_pipeline(pipeline)

        TEST_CLIENT.start_pipeline(name)

        TEST_CLIENT.push_to_pipeline(name, "tbl", "csv", data)
        resp = TEST_CLIENT.query_as_json(pipeline.name, "SELECT * FROM tbl")
        expected = [{"id": 2}, {"id": 1}]
        got = list(resp)

        self.assertCountEqual(got, expected)

        TEST_CLIENT.stop_pipeline(name, force=True)
        TEST_CLIENT.clear_storage(name)
        TEST_CLIENT.delete_pipeline(name)


if __name__ == "__main__":
    unittest.main()



import unittest

from feldera import PipelineBuilder
from tests import TEST_CLIENT
from decimal import Decimal


class TestVariant(unittest.TestCase):
    def test_local(self):
        sql = """
-- Ingest JSON as string; output it as VARIANT.
CREATE TABLE json_table (json VARCHAR) with ('materialized' = 'true');
CREATE MATERIALIZED VIEW json_view AS SELECT PARSE_JSON(json) AS json FROM json_table;
CREATE MATERIALIZED VIEW json_string_view AS SELECT TO_JSON(json) AS json FROM json_view;

CREATE MATERIALIZED VIEW average_view AS SELECT
CAST(json['name'] AS VARCHAR) as name,
((CAST(json['scores'][1] AS DECIMAL(8, 2)) + CAST(json['scores'][2] AS DECIMAL(8, 2))) / 2) as average
FROM json_view;

-- Ingest JSON as variant; extract strongly typed columns from it.
CREATE TABLE variant_table(val VARIANT) with ('materialized' = 'true');

CREATE MATERIALIZED VIEW typed_view AS SELECT
    CAST(val['name'] AS VARCHAR) as name,
    CAST(val['scores'] AS DECIMAL ARRAY) as scores
FROM variant_table;
        """

        pipeline = PipelineBuilder(
            TEST_CLIENT, name="test_variant", sql=sql
        ).create_or_replace()

        input_strings = [
            {"json": '{"name":"Bob","scores":[8,10]}'},
            {"json": '{"name":"Dunce","scores":[3,4]}'},
            {"json": '{"name":"John","scores":[9,10]}'},
        ]

        input_json = [
            {"val": {"name": "Bob", "scores": [8, 10]}},
            {"val": {"name": "Dunce", "scores": [3, 4]}},
            {"val": {"name": "John", "scores": [9, 10]}},
        ]

        expected_strings = [j | {"insert_delete": 1} for j in input_strings]

        expected_average = [
            {"name": "Bob", "average": Decimal(9)},
            {"name": "Dunce", "average": Decimal(3.5)},
            {"name": "John", "average": Decimal(9.5)},
        ]
        for datum in expected_average:
            datum.update({"insert_delete": 1})

        expected_typed = [
            {"name": "Bob", "scores": [8, 10]},
            {"name": "Dunce", "scores": [3, 4]},
            {"name": "John", "scores": [9, 10]},
        ]
        for datum in expected_typed:
            datum.update({"insert_delete": 1})

        expected_variant = [
            {"json": {"name": "Bob", "scores": [8, 10]}},
            {"json": {"name": "Dunce", "scores": [3, 4]}},
            {"json": {"name": "John", "scores": [9, 10]}},
        ]
        for datum in expected_variant:
            datum.update({"insert_delete": 1})

        variant_out = pipeline.listen("json_view")
        json_out = pipeline.listen("json_string_view")
        average_out = pipeline.listen("average_view")
        typed_out = pipeline.listen("typed_view")

        pipeline.start()

        # Feed JSON as strings, receive output from `average_view` and `json_view`
        pipeline.input_json("json_table", input_strings)

        pipeline.wait_for_completion(False)

        assert expected_average == average_out.to_dict()
        assert expected_variant == variant_out.to_dict()

        assert expected_strings == json_out.to_dict()

        # Feed VARIANT, read strongly typed columns. Since output colums have the same
        # shape as inputs, output and input should be identical.
        pipeline.input_json("variant_table", input_json)
        pipeline.wait_for_completion(False)

        assert expected_typed == typed_out.to_dict()

        pipeline.wait_for_completion(True)

        pipeline.delete(True)


if __name__ == "__main__":
    unittest.main()



from decimal import Decimal
import unittest

from pandas import Timedelta, Timestamp
from feldera import PipelineBuilder
from tests import TEST_CLIENT


class TestUDF(unittest.TestCase):
    def test_local(self):
        sql = """
CREATE TYPE my_struct AS (
   i INT,
   s VARCHAR
);

CREATE TABLE t (
    i INT,
    ti TINYINT,
    si SMALLINT,
    bi BIGINT,
    r REAL,
    d DOUBLE,
    bin VARBINARY,
    dt DATE,
    t TIME,
    ts TIMESTAMP,
    a INT ARRAY,
    m MAP<VARCHAR, VARCHAR>,
    v VARIANT,
    b BOOLEAN,
    dc DECIMAL(7,2),
    s VARCHAR,
    ms MY_STRUCT
) with ('materialized' = 'true');

CREATE FUNCTION bool2bool(i BOOLEAN) RETURNS BOOLEAN;
CREATE FUNCTION nbool2nbool(i BOOLEAN NOT NULL) RETURNS BOOLEAN NOT NULL;

CREATE FUNCTION i2i(i INT) RETURNS INT;
CREATE FUNCTION ni2ni(i INT NOT NULL) RETURNS INT NOT NULL;

CREATE FUNCTION ti2ti(i TINYINT) RETURNS TINYINT;
CREATE FUNCTION nti2nti(i TINYINT NOT NULL) RETURNS TINYINT NOT NULL;

CREATE FUNCTION si2si(i SMALLINT) RETURNS SMALLINT;
CREATE FUNCTION nsi2nsi(i SMALLINT NOT NULL) RETURNS SMALLINT NOT NULL;

CREATE FUNCTION bi2bi(i BIGINT) RETURNS BIGINT;
CREATE FUNCTION nbi2nbi(i BIGINT NOT NULL) RETURNS BIGINT NOT NULL;

CREATE FUNCTION r2r(i REAL) RETURNS REAL;
CREATE FUNCTION nr2nr(i REAL NOT NULL) RETURNS REAL NOT NULL;

CREATE FUNCTION d2d(i DOUBLE) RETURNS DOUBLE;
CREATE FUNCTION nd2nd(i DOUBLE NOT NULL) RETURNS DOUBLE NOT NULL;

CREATE FUNCTION bin2bin(i VARBINARY) RETURNS VARBINARY;
CREATE FUNCTION nbin2nbin(i VARBINARY NOT NULL) RETURNS VARBINARY NOT NULL;

CREATE FUNCTION date2date(i DATE) RETURNS DATE;
CREATE FUNCTION ndate2ndate(i DATE NOT NULL) RETURNS DATE NOT NULL;

CREATE FUNCTION ts2ts(i TIMESTAMP) RETURNS TIMESTAMP;
CREATE FUNCTION nts2nts(i TIMESTAMP NOT NULL) RETURNS TIMESTAMP NOT NULL;

CREATE FUNCTION t2t(i TIME) RETURNS TIME;
CREATE FUNCTION nt2nt(i TIME NOT NULL) RETURNS TIME NOT NULL;

CREATE FUNCTION arr2arr(i INT ARRAY) RETURNS INT ARRAY;
CREATE FUNCTION narr2narr(i INT ARRAY NOT NULL) RETURNS INT ARRAY NOT NULL;

CREATE FUNCTION map2map(i MAP<VARCHAR, VARCHAR>) RETURNS MAP<VARCHAR, VARCHAR>;
CREATE FUNCTION nmap2nmap(i MAP<VARCHAR, VARCHAR> NOT NULL) RETURNS MAP<VARCHAR, VARCHAR> NOT NULL;

CREATE FUNCTION var2var(i VARIANT) RETURNS VARIANT;
CREATE FUNCTION nvar2nvar(i VARIANT NOT NULL) RETURNS VARIANT NOT NULL;

CREATE FUNCTION dec2dec(i DECIMAL(7, 2)) RETURNS DECIMAL(7, 2);
CREATE FUNCTION ndec2ndec(i DECIMAL(7, 2) NOT NULL) RETURNS DECIMAL(7, 2) NOT NULL;

CREATE FUNCTION str2str(i VARCHAR) RETURNS VARCHAR;
CREATE FUNCTION nstr2nstr(i VARCHAR NOT NULL) RETURNS VARCHAR NOT NULL;

CREATE FUNCTION struct2struct(i my_struct) RETURNS my_struct;
CREATE FUNCTION nstruct2nstruct(i my_struct NOT NULL) RETURNS my_struct NOT NULL;

CREATE MATERIALIZED VIEW v AS
SELECT
    bool2bool(b),
    nbool2nbool(COALESCE(b, FALSE)),
    i2i(i),
    ni2ni(COALESCE(i, 0)),
    ti2ti(ti),
    nti2nti(COALESCE(ti, 0)),
    si2si(si),
    nsi2nsi(COALESCE(si, 0)),
    bi2bi(bi),
    nbi2nbi(COALESCE(bi, 0)),
    r2r(r),
    nr2nr(COALESCE(r, 0.0)),
    d2d(d),
    nd2nd(COALESCE(d, 0.0)),
    bin2bin(bin),
    nbin2nbin(COALESCE(bin, x'')),
    date2date(dt),
    ndate2ndate(COALESCE(dt, DATE '2023-01-01')),
    ts2ts(ts),
    nts2nts(COALESCE(ts, TIMESTAMP '2023-01-01 00:00:00')),
    t2t(t),
    nt2nt(COALESCE(t, TIME '00:00:00')),
    arr2arr(a),
    narr2narr(a),
    map2map(m),
    nmap2nmap(m),
    var2var(v),
    nvar2nvar(COALESCE(v, VARIANTNULL())),
    dec2dec(dc),
    ndec2ndec(COALESCE(dc, 0)),
    str2str(s),
    nstr2nstr(COALESCE(s, ''))
FROM
    t;
        """

        udfs = """
use feldera_sqllib::F32;
use crate::*;

pub fn bool2bool(i: Option<bool>) -> Result<Option<bool>, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn nbool2nbool(i: bool) -> Result<bool, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn i2i(i: Option<i32>) -> Result<Option<i32>, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn ni2ni(i: i32) -> Result<i32, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn ti2ti(i: Option<i8>) -> Result<Option<i8>, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn nti2nti(i: i8) -> Result<i8, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn si2si(i: Option<i16>) -> Result<Option<i16>, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn nsi2nsi(i: i16) -> Result<i16, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn bi2bi(i: Option<i64>) -> Result<Option<i64>, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn nbi2nbi(i: i64) -> Result<i64, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn r2r(i: Option<F32>) -> Result<Option<F32>, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn nr2nr(i: F32) -> Result<F32, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn d2d(i: Option<F64>) -> Result<Option<F64>, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn nd2nd(i: F64) -> Result<F64, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn bin2bin(i: Option<ByteArray>) -> Result<Option<ByteArray>, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn nbin2nbin(i: ByteArray) -> Result<ByteArray, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn date2date(i: Option<Date>) -> Result<Option<Date>, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn ndate2ndate(i: Date) -> Result<Date, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn ts2ts(i: Option<Timestamp>) -> Result<Option<Timestamp>, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn nts2nts(i: Timestamp) -> Result<Timestamp, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn t2t(i: Option<Time>) -> Result<Option<Time>, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn nt2nt(i: Time) -> Result<Time, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn arr2arr(i: Option<Array<Option<i32>>>) -> Result<Option<Array<Option<i32>>>, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn narr2narr(i: Array<Option<i32>>) -> Result<Array<Option<i32>>, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn map2map(i: Option<Map<SqlString, Option<SqlString>>>) -> Result<Option<Map<SqlString, Option<SqlString>>>, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn nmap2nmap(i: Map<SqlString, Option<SqlString>>) -> Result<Map<SqlString, Option<SqlString>>, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn var2var(i: Option<Variant>) -> Result<Option<Variant>, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn nvar2nvar(i: Variant) -> Result<Variant, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn dec2dec(i: Option<SqlDecimal>) -> Result<Option<SqlDecimal>, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn ndec2ndec(i: SqlDecimal) -> Result<SqlDecimal, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn str2str(i: Option<SqlString>) -> Result<Option<SqlString>, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn nstr2nstr(i: SqlString) -> Result<SqlString, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn struct2struct(i: Option<Tup2<Option<i32>, Option<SqlString>>>) -> Result<Option<Tup2<Option<i32>, Option<SqlString>>>, Box<dyn std::error::Error>> {
    Ok(i)
}
pub fn nstruct2nstruct(i: Tup2<Option<i32>, Option<SqlString>>) -> Result<Tup2<Option<i32>, Option<SqlString>>, Box<dyn std::error::Error>> {
    Ok(i)
}
        """

        pipeline = PipelineBuilder(
            TEST_CLIENT, name="test_udfs", sql=sql, udf_rust=udfs
        ).create_or_replace()

        # TODO: use .query() instead
        out = pipeline.listen("v")

        pipeline.start()

        pipeline.input_json(
            "t",
            [
                {
                    "i": 1,
                    "ti": -2,
                    "si": 3,
                    "bi": -4,
                    "r": 0.5,
                    "d": 1e-5,
                    "bin": [],
                    "dt": "2024-09-25",
                    "t": "13:05:00",
                    "ts": "2024-09-25 13:05:00",
                    "a": [1, 2, 3, 4, 5],
                    "m": {"foo": "bar"},
                    "v": '{"foo": "bar"}',
                    "b": True,
                    "dc": "123.45",
                    "s": "foobar",
                }
            ],
        )
        pipeline.wait_for_idle()

        output = out.to_dict()
        assert output == [
            {
                "EXPR$0": True,
                "EXPR$1": True,
                "EXPR$10": 0.5,
                "EXPR$11": 0.5,
                "EXPR$12": 1e-05,
                "EXPR$13": 1e-05,
                "EXPR$14": [],
                "EXPR$15": [],
                "EXPR$16": Timestamp("2024-09-25 00:00:00"),
                "EXPR$17": Timestamp("2024-09-25 00:00:00"),
                "EXPR$18": Timestamp("2024-09-25 13:05:00"),
                "EXPR$19": Timestamp("2024-09-25 13:05:00"),
                "EXPR$2": 1,
                "EXPR$20": Timedelta("0 days 13:05:00"),
                "EXPR$21": Timedelta("0 days 13:05:00"),
                "EXPR$22": [1, 2, 3, 4, 5],
                "EXPR$23": [1, 2, 3, 4, 5],
                "EXPR$24": {"foo": "bar"},
                "EXPR$25": {"foo": "bar"},
                "EXPR$26": '{"foo": "bar"}',
                "EXPR$27": '{"foo": "bar"}',
                "EXPR$28": Decimal("123.45"),
                "EXPR$29": Decimal("123.45"),
                "EXPR$3": 1,
                "EXPR$30": "foobar",
                "EXPR$31": "foobar",
                "EXPR$4": -2,
                "EXPR$5": -2,
                "EXPR$6": 3,
                "EXPR$7": 3,
                "EXPR$8": -4,
                "EXPR$9": -4,
                "insert_delete": 1,
            }
        ]

        pipeline.stop(force=True)
        pipeline.delete(True)


if __name__ == "__main__":
    unittest.main()
